/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#ifndef ARCH_HPP_
#define ARCH_HPP_

#include <iostream>
#include <vector>
#include "tile.hpp"
#include "app.hpp"
#include "misc.hpp"

typedef class Link{
    friend ostream & operator<<(ostream &os, const Link & link);
    static int cnt;
    float powerCoeff;             // Average energy consumption of sending one bit data
                                  // over that link
    int id;
    Position fromTile;
    Position toTile;
    int fromTileId;
    int toTileId;
    int used_BW;

public:
    int bandwidth;
    float Cost() {return powerCoeff;}
    int GetId() {return id;}
    Link();
    int const GetCnt();
    const Position & FromTilePos();
    const Position & ToTilePos();
    int FromTile() { return fromTileId;}
    int ToTile() {return toTileId;}
    int direction();
    friend bool verify_BW_requirement();
    friend int locate_link(Position & pos, int direction);
}*pLink, Link;

typedef class Tile{
    static int cnt;
    static int totalTiles;
    //    static int rowNum;
    int id;
    int switch_delay;        //The delay of the switch in the tile (in cycles unit)
    int arbitration_delay;   //The delay of the routing. In wormhole routing, only
                             //processing the header flit will incur this delay
    float powerCoeff;        //The average energy consumption of routing one bit data
    Position pos;
    int goLinkNum;
    int comeLinkNum;
    vector<pLink> goLink;
    vector<pLink> comeLink;
    int procId;

    class Router{
        ROUTING_TYPE routing_type;
        Tile * host_tile;               
        int **routing_table;      //routing_table[i][j] shows how to which link to send the 
        //from tile i to tile j. If it's -2, means not reachable.
        //if it's -1, means the destination is the current tile.
    
        bool generate_xy_routing_table();
    public:
        Router();
        bool initialize(ROUTING_TYPE type, Tile * host);
        int route_to_link(int src_tile, int dst_tile) const;
        int set_routing_entry(int src_tile, int dst_tile, int link_id);
        ~Router();
        int choose_egress(int src_tile, int dst_tile);
    } router;

public:
    Tile();
    float Cost() {return powerCoeff;}
    int mapToProc() { return procId;}
    int mapToProc(int pId) { procId = pId; return pId;}
    int GetId() const;
    int GetGoLinkNum() const;
    int GetComeLinkNum() const;

    bool initialize_router(ROUTING_TYPE routing_type);
    int RouteToLink(int srcId, int dstId) const;
    int RouteToLink(const Tile& srcTile, const Tile& dstTile) const;
    int RouteToLink(int srcId, int dstId, int linkId);
    int set_routing_entry(int src_tile, int dst_tile, int link_id);
    Position GetPosition() const;
    pLink GoLink(int i);
    pLink ComeLink(int i);
    void AttachLink(pLink gLink);
    ~Tile();
    friend ostream & operator<<(ostream & os, const Tile & tile);
}*pTile, Tile;

/************************************************************
findLink is used to find out the id of the link that connects
the source tile with the destination tile. If no link exists, 
return -1
************************************************************/
int findLink(int srcTile, int dstTile);

#endif //ARCH_HPP_
