/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */


#ifndef UTIL_HPP_
#define UTIL_HPP_

#include <fstream>
#include <cassert>
#include "tile.hpp"
#include "gVal.hpp"
#include "app.hpp"
#include "arch.hpp"

#define SEQUENTIAL_MAPPING 0
#define RANDOM_MAPPING 1

ostream & operator<<(ostream & os, const Position & pos);
ostream & operator<<(ostream & os, const Position3D & pos);
int toAbsAddr(const Position & pos);
int toAbsAddr(const Position3D & pos);

bool parse_options(int argn, char ** argv);

bool initialize(void);
void clean();

//Free or build the link usage matrix
//The matrix is three dimensional. 
//If g_link_usage_matrix[tile1][tile2][linkId] = 1 means the 
//communication uses the corresponding link. Otherwise, the 
//link is not used for the communication.
void delete_link_usage_matrix();
void delete_link_usage_list();
void build_link_usage_matrix();

bool clearRoutingTables(void);

bool operator==(const Position & pos1, const Position &pos2);
bool operator!=(const Position & pos1, const Position &pos2);
bool operator==(const Position3D & pos1, const Position3D & pos2);
bool operator!=(const Position3D & pos1, const Position3D & pos2);

/***********************************************************
testRoutability is used to test whether packets from srcTile 
is routable to dstTile.
return TRUE: routable
       FALSE: Not routable.
************************************************************/
bool testRoutability(Tile& srcTile, Tile& dstTile);

/************************************************************
Function verify() is used to verify whether the chip's 
communication architecture can work correctly
************************************************************/
bool verify();

/************************************************************
This function is used to verify whether each link's 
communication load exceeds its bandwidth capability
************************************************************/
bool verify_BW_requirement();

/************************************************************
Function mapProcToFile maps processes to the tiles.
type: SEQUENTIAL_MAPPING
      RANDOM_MAPPING
return TRUE:  mapping succeed
       FALSE: mapping failed
************************************************************/
bool mapProcToTile(int type);

void padWriteFile(ofstream & oFile, int d1, int d2, int d3);
void padWriteFile(ofstream & oFile, int lId);

void print_help();

int peakComm();
int totalComm();

float calc_comm_energy();
float calc_switch_energy();
float calc_link_energy();
float calc_buffer_energy();

void synthesizeRoutingTable();

void analyzeIt();

void dispMapping();

void printTimes(clock_t real, struct tms *tmsstart, struct tms *tmsend);

bool dump_mapping_table(char* file_name);
bool read_mapping_table(char* file_name);
bool dump_apcg(char * file_name);

int locate_link(Position & pos, int direction);

bool exist_locked_pe(void);
bool exist_non_regular_regions(void);

/**********************************************************************
This function generate array_size numbers between min and max, all the 
numbers are guaranteed to be different
**********************************************************************/ 
int * generate_random_array(int array_size, int min, int max);

void print_comm_statistics();

bool dump_routing_tables(char * file_name);

#endif //UTIL_HPP_



