/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */


#ifndef POWER_HPP_
#define POWER_HPP_

#include <vector>
#include <iostream>
#include <fstream>
#include "common.hpp"

#ifdef USE_ORION
extern "C" {
#include "SIM_power.h"
#include "SIM_router_power.h"
#include "SIM_power_router.h"
#include "SIM_power_ALU.h"
}

typedef class Power_module {
    SIM_power_router_info_t router_info_;
    SIM_power_router_t router_power_;
    SIM_power_arbiter_t arbiter_power_;
    SIM_power_bus_t link_power_;
//     SIM_power_ALU_t alu_power_dir_x_;      
//     SIM_power_ALU_t alu_power_dir_y_;
//     SIM_power_ALU_t alu_power_buf_cmp_;      

    // data/status from the last transaction
    vector<Atom_type> buffer_write_;
    vector<Atom_type> crossbar_read_;
    vector<Atom_type> crossbar_write_;
    vector<Atom_type> link_traversal_;    
    vector<int> crossbar_input_;
    vector<Atom_type> arbiter_req_;
    vector<unsigned int> arbiter_grant_;

public:
    Power_module(int a);
    void power_buffer_read(int in_port, Atom_type read_d);
    void power_buffer_write(int in_port, Atom_type write_d);
    void power_crossbar_trav(int in_port, int out_port, Atom_type trav_d);
    void power_arbiter(int pc, Atom_type req, unsigned int gra);
    void power_link_traversal(int in_port, Atom_type read_d);
    double power_buffer_report();
    double power_link_report();
    double power_crossbar_report();
    double power_arbiter_report();
//     void power_alu_x(int pc, Atom_type d1, Atom_type d2);
//     void power_alu_y(int pc, Atom_type d1, Atom_type d2);
//     void power_alu_buf(int pc, Atom_type d1, Atom_type d2);
//     double power_alu_report();
};

typedef Power_module *pPower_module;
#endif  // USE_ORION

#endif // POWER_HPP_
