/*
 * Copyright (c) 2005 Carnegie Mellon University.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Carnegie Mellon University (http://www.cmu.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Carnegie Mellon University" and "CMU" must
 *    not be used to endorse or promote products derived from this
 *    software.
 *
 * 5. Products derived from this software may not be called
 *     "Carnegie Mellon University" or "CMU",
 *    nor may "Carnegie Mellon University"  or "CMU" appear in their name.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @ProposedRating yellow jingcao@ece.cmu.edu
 */

#ifndef ROUTER_HPP_
#define ROUTER_HPP_

#include <vector>
#include "common.hpp"
#include "channel.hpp"
#include "routing_engine.hpp"
#include "switch_fabric.hpp"
#include "arbiter.hpp"
#include "network.hpp"
#include "port.hpp"
#include "power.hpp"

typedef class Router : public Addressee {
private:
    class Network * network;

    vector<class In_port *> in_ports;
    vector<class Out_port *> out_ports;

    class Switch_fabric sw;
    class Arbiter arbiter;

    class Power_module * power_module;    // module for orion power model

public:
    Router(class Network * n, struct Position p);
    ~Router();
    class Network * get_network(void) { return network; }
    class In_port * get_in_port(Direction dir) { return in_ports[dir]; }
    class Out_port * get_out_port(Direction dir) { return out_ports[dir]; }
    class Input_channel * get_in_channel(Direction dir, int index = 0);
    class Output_channel * get_out_channel(Direction dir, int index = 0);
    class Router * get_router(Direction dir);

    class Arbiter * get_arbiter(void) { return &arbiter; }
    class Switch_fabric * get_switch(void) { return &sw; }
    void set_switch_ebit(double e) { sw.set_ebit(e); }

    void set_input_channel_buffer_size(int s);
    void set_input_channel_buffer_size(int s, Direction dir);
    void set_output_channel_buffer_size(int s);
    void set_routing_scheme(Routing_scheme rs);
    void set_routing_table(char *table_string); // UYO

    class Power_module * get_power_module(void) { return power_module; }

    void in_ports_tick(void);
    void out_ports_tick(void);
    void arbiter_tick();
};
typedef Router *pRouter;


#endif  // ROUTER_HPP_
